from ewah.hooks.base import EWAHBaseHook

from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.adsinsights import AdsInsights

from datetime import datetime, timedelta

import time


class EWAHFacebookHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "app_id": "login",
        "app_secret": "password",
        "api_version": "host",
        "account_id": "schema",
    }

    conn_name_attr = "ewah_facebook_conn_id"
    default_conn_name = "ewah_facebook_default"
    conn_type = "ewah_facebook"
    hook_name = "EWAH Facebook Connection"

    _DEFAULT_VERSION = "v13.0"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "extra"],
            "relabeling": {
                "host": "[Optional] API Version",
                "schema": "[Optional] Account ID",
                "login": "App ID",
                "password": "App Secret",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget

        return {
            "extra__ewah_facebook__access_token": StringField(
                "Access Token", widget=BS3PasswordFieldWidget()
            ),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fb_init(self, account_id=None):
        account_id = account_id or self.conn.account_id or None
        if account_id:
            account_id = str(account_id)
            if not account_id.startswith("act_"):
                account_id = "act_{0}".format(account_id)
        api_version = self.conn.api_version or self._DEFAULT_VERSION
        if not api_version.startswith("v"):
            api_version = "v{0}".format(api_version)
        FacebookAdsApi.init(
            app_id=self.conn.app_id or None,
            app_secret=self.conn.app_secret or None,
            access_token=self.conn.access_token,
            api_version=api_version,
            account_id=account_id,
        )

    def get_data_in_batches(
        self,
        level,
        fields,
        data_from,
        data_until,
        breakdowns=None,
        account_id=None,
        batch_size=10000,
    ):
        if hasattr(data_from, "date"):
            data_from = data_from.date()
        if hasattr(data_until, "date"):
            data_until = data_until.date()
        self.fb_init(account_id=account_id)
        account_id = str(account_id or self.conn.account_id)
        if not account_id.startswith("act_"):
            account_id = "act_{0}".format(account_id)
        account_object = AdAccount(account_id)

        params = {
            "time_increment": 1,
            "level": level,
            "limit": 1000,
        }
        if breakdowns:
            if isinstance(breakdowns, list):
                breakdowns = ",".join(breakdowns)
            params["breakdowns"] = breakdowns

        data = []
        # maximum request of 90 days at once!
        while data_from <= data_until:
            request_until = min(data_until, data_from + timedelta(days=90))
            params["time_range"] = {
                "since": data_from.strftime("%Y-%m-%d"),
                "until": request_until.strftime("%Y-%m-%d"),
            }
            data_from = request_until + timedelta(days=1)
            self.log.info(
                "Requesting data for account_id={0} between {1} and {2}.".format(
                    str(account_id),
                    params["time_range"]["since"],
                    params["time_range"]["until"],
                )
            )
            async_job = account_object.get_insights_async(
                fields=fields,
                params=params,
            )

            job_read = async_job.api_get()
            while not job_read.get("async_status") in [
                "Job Completed",
                "Job Failed",
                "Job Skipped",
            ]:
                self.log.info(
                    "Asnyc job completion: {0}% (status: {1})".format(
                        str(job_read.get("async_percent_completion")),
                        str(job_read.get("async_status")),
                    )
                )
                time.sleep(5)
                job_read = async_job.api_get()

            time.sleep(1)
            assert (
                job_read.get("async_status") == "Job Completed"
            ), job_read.get_result()

            data += [
                {
                    k: datetime.strptime(v, "%Y-%m-%d").date()
                    if k in ("date_start", "date_stop")
                    else v
                    for k, v in datum.items()
                }
                for datum in list(job_read.get_result(params={"limit": 1000}))
            ]

            if len(data) >= batch_size:
                yield data
                data = []

        if data:
            yield data



from airflow.models import Variable
from airflow.configuration import conf

from ewah.hooks.base import EWAHBaseHook

import requests
import json
from datetime import datetime, timedelta
from selenium import webdriver
from cryptography.fernet import Fernet


class EWAHPlentyMarketsHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "username": "login",
        "url": "host",
    }

    conn_name_attr = "ewah_plentymarkets_conn_id"
    default_conn_name = "ewah_plentymarkets_default"
    conn_type = "ewah_plentymarkets"
    hook_name = "EWAH PlentyMarkets Connection"

    _INCREMENTAL_FIELDS = {
        "/rest/orders": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/status-history": ["createdAtFrom", "createdAtTo"],
        "/rest/accounts/contacts": ["updatedAtAfter", "updatedAtBefore"],
        # TODO - refactor this hook to not need this dict to begin with
        "/rest/orders/documents/invoice": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/receipt": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/tillCount": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/posCouponReceipt": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/posInvoice": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/posInvoiceCancellation": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/cancellation": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/zReport": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/invoiceExternal": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/invoice": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/deliveryNote": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/poDeliveryNote": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/creditNote": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/creditNoteExternal": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/orderConfirmation": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/offer": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/dunningLetter": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/reversalDunningLetter": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/returnNote": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/successConfirmation": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/correction": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/reversal": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/reversalRefund": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/proFormaInvoice": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/reorder": ["updatedAtFrom", "updatedAtTo"],
        "/rest/orders/documents/uploaded": ["updatedAtFrom", "updatedAtTo"],
    }

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra"],
            "relabeling": {
                "login": "Username",
                "password": "Password",
                "host": "URL (e.g. https://myplentyshop.com or https://plentymarkets-cloud-07.com/12345)",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget

        return {
            "extra__ewah_plentymarkets__url_is_final": StringField(
                "Is this the final endpoint?",
                default="no",
                widget=BS3TextFieldWidget(),
            )
        }

    @property
    def endpoint(self):
        if self.conn.url_is_final.lower().startswith(("y", "t")):
            return self.conn.url

        # get the current endpoint with the correct backend hash
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1420,1080")
        driver = webdriver.Chrome(options=chrome_options)
        driver.get(self.conn.url)
        driver.find_element_by_id("username").send_keys(self.conn.username)
        driver.find_element_by_id("password").send_keys(self.conn.password)
        driver.find_element_by_xpath('//button[normalize-space()="Login"]').click()
        endpoint = driver.current_url
        if endpoint.count("/") > 2:
            endpoint = endpoint[
                : endpoint.find("/", endpoint.find("/", endpoint.find("/") + 1) + 1)
            ]
        return endpoint

    @property
    def token(self):
        if not hasattr(self, "_token") or datetime.now() > self._token_expires_at:
            # attempt to get the token from the Variable (if stored there)
            # Variable is encrypted with airflow instance's own fernet key
            # In the rare event of a key change, just load a new token
            # If currently rotating keys, use the first one
            fernet = Fernet(conf.get("core", "fernet_key").split(".")[0])
            ewah_plenty_token_variable_key = "__ewah_plenty_token"
            try:
                variable_decrypted = json.loads(
                    fernet.decrypt(
                        Variable.get(
                            key=ewah_plenty_token_variable_key, default_var=None
                        ).encode()
                    ).decode()
                )
            except:
                variable_decrypted = None

            if (
                variable_decrypted
                and variable_decrypted.get("expires_at")
                and datetime.fromisoformat(variable_decrypted["expires_at"])
                > datetime.utcnow()
            ):
                self.log.info("Using token stored as airflow variable")
                self._token = variable_decrypted["access_token"]
                self._token_expires_at = datetime.fromisoformat(
                    variable_decrypted["expires_at"]
                )
            else:
                self.log.info("Requesting new token")
                requested_at = datetime.now()
                token_request = requests.post(
                    self.endpoint + "/rest/login",
                    params={
                        "username": self.conn.username,
                        "password": self.conn.password,
                    },
                    headers={
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                    },
                )
                assert token_request.status_code == 200, token_request.text
                try:
                    request_data = token_request.json()
                except:
                    assert False, "Response is not a JSON - Response Text: {0}".format(
                        token_request.text
                    )
                self._token_expires_at = requested_at + timedelta(
                    seconds=request_data["expires_in"]
                )
                self._token = request_data["access_token"]
                Variable.set(
                    key=ewah_plenty_token_variable_key,
                    value=fernet.encrypt(
                        json.dumps(
                            {
                                "expires_at": self._token_expires_at.isoformat(),
                                "access_token": self._token,
                            }
                        ).encode()
                    ).decode(),
                )
        return self._token

    @staticmethod
    def format_resource(resource):
        if "rest/" in resource:
            if not resource.startswith("/"):
                resource = "/" + resource
        else:
            resource = "/rest/{0}".format(resource)
        return resource

    def get_data_in_batches(
        self,
        resource,
        data_from=None,
        data_until=None,
        batch_size=10000,
        additional_params=None,
    ):
        params = {
            "itemsPerPage": 250,  # Maximum
        }
        if additional_params:
            assert isinstance(additional_params, dict)
            params.update(additional_params)

        resource = self.format_resource(resource)
        url = self.endpoint + resource

        if data_from:
            params[self._INCREMENTAL_FIELDS[resource][0]] = data_from.isoformat()
        if data_until:
            if resource == "/rest/accounts/contacts":
                # inconsistent API implementation - ignores data for last day otherwise
                data_until += timedelta(days=1)
            params[self._INCREMENTAL_FIELDS[resource][1]] = data_until.isoformat()

        data = []
        while True:
            headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "Authorization": "Bearer {0}".format(self.token),
            }
            self.log.info("Requesting new page of data...")
            data_request = requests.get(url, params=params, headers=headers)
            assert data_request.status_code in (200, 401), "Status {0}: {1}".format(
                data_request.status_code, data_request.text
            )
            if data_request.status_code == 401:
                # Sometimes, the backend hash changes. If this happens during
                # a data load, is will result in a specific 401 error. Catch
                # the error and try again (once) with the new endpoint. The
                # try is needed in case the data_request() does not return
                # a JSON (will produce an error otherwise).
                try:
                    if data_request.json()["class"] == "UIHashExpiredException":
                        if "rest/" in resource:
                            url = self.endpoint + resource
                        else:
                            url = self.endpoint + "/rest/{0}".format(resource)
                        data_request = requests.get(url, params=params, headers=headers)
                except:
                    pass  # assert below will take care of any error
            assert data_request.status_code == 200, "Status {0}: {1}".format(
                data_request.status_code, data_request.text
            )
            returned_data = data_request.json()
            if isinstance(returned_data, list):
                # some endpoints return just one page with the data as list
                data += returned_data
                break
            else:
                if "entries" in returned_data.keys():
                    data += returned_data["entries"]
                else:
                    # Special case where just a dict with key-value pairs is returned
                    # Example: /rest/accounts/contacts/classes
                    data += [{"id": k, "value": v} for k, v in returned_data.items()]
                    break
                params["page"] = returned_data["page"] + 1  # for next request
                if returned_data["isLastPage"]:
                    break
            if len(data) >= batch_size:
                yield data
                data = []

        if data:
            yield data



from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook
from ewah.uploaders.base import EWAHBaseUploader

from airflow.operators.python import PythonOperator
from airflow.models import BaseOperator

from google.cloud.bigquery.table import TableReference
from google.cloud.bigquery import (
    Table,
    SchemaField,
    LoadJobConfig,
    CopyJobConfig,
)

from copy import deepcopy
from time import sleep
from tempfile import TemporaryFile, TemporaryDirectory
from datetime import datetime, date, timedelta

from avro.datafile import DataFileWriter
from avro.io import DatumWriter

import avro.schema
import json
import os


class BigqueryOperator(BaseOperator):
    "Operate to execute SQL on BigQuery"

    def __init__(
        self, sql, bigquery_conn_id, project=None, params=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sql = sql
        self.conn_id = bigquery_conn_id
        self.project = project
        # an attribute called "params" of an operator leads to an airflow error
        # if that attribute is None due to default templating of "task.params"!
        self._params = params

    def execute(self, context):
        from ewah.hooks.bigquery import EWAHBigQueryHook

        hook = EWAHBigQueryHook(
            conn_id=self.conn_id,
            project_id=self.project,
        )
        hook.execute(self.sql, commit=True, params=self._params)
        hook.commit()
        hook.close()


class FakeDatasetRef:
    def __init__(self, dataset_id, project_id):
        self.dataset_id = dataset_id
        self.project = project_id


def map_bq_data_type_to_avro(data_type):
    return {
        "STRING": "string",
        "INT64": "long",
        "BOOL": "boolean",
        # "DATE": "", TODO: Properly deal with datetime types
        # "TIMESTAMP": "",
        "BYTES": "bytes",
        "FLOAT64": "double",
    }[data_type]


class EWAHBigQueryUploader(EWAHBaseUploader):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT column_name
        FROM `{project_id}.{schema_name}.INFORMATION_SCHEMA.COLUMNS`
        WHERE table_name = '{table_name}'
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE `{project_id}.{schema_name}.{table_name}`
        ADD COLUMN `{column_name}` {column_type};
    """

    _QUERY_TABLE = "SELECT * FROM `{project_id}.{schema_name}.{table_name}`"

    def __init__(
        self,
        *args,
        partition_field=None,
        partition_type=None,
        require_partition_filter=False,
        insert_chunk_size=100,
        **kwargs,
    ) -> None:
        super().__init__(EC.DWH_ENGINE_BIGQUERY, *args, **kwargs)
        # BigQuery project id aka database name may be set in connection
        self.database_name = self.database_name or self.dwh_hook.conn.project
        if partition_type or partition_field or require_partition_filter:
            assert (
                partition_type and partition_field
            ), "partition_type and partition_field must both be set if either is set or require_partition_filter is true!"
        self.partition_field = partition_field
        self.partition_type = partition_type
        self.insert_chunk_size = insert_chunk_size

        assert self.use_temp_pickling, "BigQuery operator only works with pickling!"

    @classmethod
    def get_cleaner_callables(cls):
        def bigquery_data_adjustments(row):
            key_changes = []
            for key, value in row.items():
                # change date, datetime, timedelta types
                # reason: non-json-serializable types are not accepted
                if isinstance(value, (date, datetime)):
                    row[key] = value.isoformat()
                elif isinstance(value, timedelta):
                    row[key] = value.total_seconds()

                # prefix field names that start with a number because BigQuery
                # does not like those types of field names.
                # Sub-challenge: generally prefix all field names that are hex
                try:
                    int(key, 16)
                    # if that did not fail, it is a hex!
                    key_changes.append((key, "field_" + key))
                except ValueError:
                    # not a hex, does it start with a number still?
                    if key[:1].isdigit():
                        key_changes.append((key, "field_" + key))

            # execute key changes
            for old_key, new_key in key_changes:
                row[new_key] = row.pop(old_key)

            return row

        return [bigquery_data_adjustments]

    @classmethod
    def get_schema_tasks(
        cls,
        dag,
        dwh_engine,
        dwh_conn_id,
        target_schema_name,
        target_schema_suffix="_next",
        target_database_name=None,
        read_right_users=None,  # Only for PostgreSQL
        **additional_task_args,
    ):
        schema_full_path = ""
        if target_database_name:
            schema_full_path += target_database_name + "."
        schema_full_path += target_schema_name

        # Create functions that are subsequently used in a PythonOperator

        def kickoff_func(schema_full, schema_suffix, dwh_conn_id):
            # kickoff: create new dataset
            schema = schema_full + schema_suffix
            conn = EWAHBaseHook.get_hook_from_conn_id(dwh_conn_id).dbconn
            # delete dataset first if it already exists
            print("Deleting the dataset {0} if it already exists.".format(schema))
            conn.delete_dataset(schema, delete_contents=True, not_found_ok=True)
            print("Creating the dataset {0}.".format(schema))
            conn.create_dataset(schema)
            print("Done!")

        def final_func(schema_name, schema_suffix, dwh_conn_id):
            # final: move new data into the final dataset
            conn = EWAHBaseHook.get_hook_from_conn_id(dwh_conn_id).dbconn
            # get dataset objects
            try:  # create final dataset if not exists
                ds_final = conn.get_dataset(schema_name)
            except:
                print("Creating dataset {0}".format(schema_name))
                ds_final = conn.create_dataset(schema_name)
            ds_temp = conn.get_dataset(schema_name + schema_suffix)

            # copy all tables from temp dataset to final dataset
            new_tables = conn.list_tables(ds_temp)
            new_table_ids = [table.table_id for table in conn.list_tables(ds_temp)]
            old_table_ids = [table.table_id for table in conn.list_tables(ds_final)]
            copy_jobs = []
            for table in new_tables:
                print(
                    "Copying table {0} from temp to final dataset".format(
                        table.table_id
                    )
                )
                try:
                    old_table = conn.get_table(
                        table=TableReference(
                            dataset_ref=ds_final, table_id=table.table_id
                        )
                    )
                    conn.delete_table(old_table)
                except:
                    # ignore failure, fails if old table does not exist to begin with
                    pass
                finally:
                    final_table = ds_final.table(table.table_id)
                    copy_jobs.append(conn.copy_table(table, final_table))

            # delete tables that don't exist in temp dataset from final dataset
            for table_id in old_table_ids:
                if not table_id in new_table_ids:
                    print("Deleting table {0}".format(table_id))
                    conn.delete_table(
                        conn.get_table(
                            TableReference(dataset_ref=ds_final, table_id=table_id)
                        )
                    )

            # make sure all copy jobs succeeded
            while copy_jobs:
                sleep(0.1)
                job = copy_jobs.pop(0)
                job.result()
                assert job.state in ("RUNNING", "DONE")
                if job.state == "RUNNING":
                    copy_jobs.append(job)
                else:
                    print(
                        "Successfully copied {0}".format(
                            job.__dict__["_properties"]["configuration"]["copy"][
                                "destinationTable"
                            ]["tableId"]
                        )
                    )

            # delete temp dataset
            print("Deleting temp dataset.")
            conn.delete_dataset(ds_temp, delete_contents=True, not_found_ok=False)

            print("Done.")

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update(
            {
                "task_id": "kickoff",
                "dag": dag,
                "python_callable": kickoff_func,
                "op_kwargs": {
                    "schema_full": schema_full_path,
                    "schema_suffix": target_schema_suffix,
                    "dwh_conn_id": dwh_conn_id,
                },
            }
        )
        task_2_args.update(
            {
                "task_id": "final",
                "dag": dag,
                "python_callable": final_func,
                "op_kwargs": {
                    "schema_name": target_schema_name,
                    "schema_suffix": target_schema_suffix,
                    "dwh_conn_id": dwh_conn_id,
                },
            }
        )

        return (PythonOperator(**task_1_args), PythonOperator(**task_2_args))

    def rollback(self):
        raise Exception("Rollback attempted - BigQuery knows no transactions!")

    def close(self):
        pass  # nothing to do

    def test_if_table_exists(self, table_name, schema_name, project_id=None):
        if project_id:
            dataset_path = "{0}.{1}".format(project_id, schema_name)
        else:
            dataset_path = schema_name
        if not self.test_if_dataset_exists(dataset_path):
            return False
        return 0 < len(
            self.dwh_hook.execute_and_return_result(
                sql="""
                        SELECT * FROM `{dataset}.__TABLES__`
                        WHERE table_id = '{table_name}'
                    """.format(
                    dataset=dataset_path, table_name=table_name
                )
            )
        )

    def test_if_dataset_exists(self, dataset_id):
        conn = self.dwh_hook.dbconn
        try:
            ds = conn.get_dataset(dataset_id)
        except:
            # fails if dataset does not exist
            return False
        return True

    def get_max_value_of_column(self, column_name):
        return self.dwh_hook.execute_and_return_result(
            sql="SELECT MAX(`{0}`) FROM `{1}.{2}.{3}`".format(
                column_name,
                self.database_name,
                self.schema_name + self.schema_suffix,
                self.table_name,
            ),
            return_dict=False,
        )[0][0]

    def drop_table_if_exists(self, table_name, schema_name, project_id=None):
        if self.test_if_table_exists(
            table_name=table_name,
            schema_name=schema_name,
            project_id=project_id,
        ):
            conn = self.dwh_hook.dbconn
            conn.delete_table(
                conn.get_table(
                    TableReference(
                        dataset_ref=FakeDatasetRef(
                            dataset_id=schema_name,
                            project_id=project_id or self.database_name,
                        ),
                        table_id=table_name,
                    )
                )
            )

    def copy_table(self):
        # Overwrite parent method for alternative approach
        # If target table was created by a previous run, drop it first
        self.drop_table_if_exists(
            self.table_name, self.schema_name + self.schema_suffix, self.database_name
        )
        # Create target table if source table exists
        if self.test_if_table_exists(
            table_name=self.table_name,
            schema_name=self.schema_name,
            project_id=self.database_name,
        ):
            self.log.info("Copying table into temporary dataset!")
            conn = self.dwh_hook.dbconn
            ds_old = conn.get_dataset(self.schema_name)
            ds_new = conn.get_dataset(self.schema_name + self.schema_suffix)
            table_old = conn.get_table(
                table=TableReference(dataset_ref=ds_old, table_id=self.table_name)
            )
            table_new = ds_new.table(self.table_name)
            copy_job = conn.copy_table(
                table_old,
                table_new,
                job_config=CopyJobConfig(write_disposition="WRITE_TRUNCATE"),
            )

            copy_job.result()  # Waits until the job is done
            assert copy_job.state == "DONE", "Unexpected job state: {0}".format(
                job.state
            )
            self.log.info(
                "Successfully copied {0}!".format(
                    copy_job.__dict__["_properties"]["configuration"]["copy"][
                        "destinationTable"
                    ]["tableId"]
                )
            )

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        columns_definition,
        load_strategy,
        upload_call_count,
        database_name=None,
        primary_key=None,
    ):
        # This method doesn't actually create or update a table. It just creates
        # and populates a single .avro file which is used in the data upload.
        # The actual upload happens when the commit() method is called.
        if upload_call_count == 1:
            # Create avro writer and file in temporary folder
            self.avro_folder = TemporaryDirectory()
            self.avro_file_name = self.avro_folder.name + os.sep + table_name + ".avro"
            avro_schema = avro.schema.parse(
                json.dumps(
                    {
                        "type": "record",
                        "name": table_name,
                        "namespace": table_name,
                        "fields": [
                            {
                                "name": name,
                                "type": [
                                    "null",
                                    map_bq_data_type_to_avro(field["data_type"]),
                                ],
                            }
                            for name, field in columns_definition.items()
                        ],
                    }
                )
            )
            # Create the avro_writer object to be used going forward
            self.avro_writer = DataFileWriter(
                open(self.avro_file_name, "wb"), DatumWriter(), avro_schema
            )
            # Save the relevant kwargs for later use in the commit() method
            self.table_creation_config = {
                "table_name": table_name,
                "schema_name": schema_name,
                "schema_suffix": schema_suffix,
                "columns_definition": columns_definition,
                "load_strategy": load_strategy,
                "database_name": database_name,
                "primary_key": primary_key,
            }

        self.log.info(
            "BigQuery Uploader writes data into Avro file for later one-off upload!"
        )
        while data:
            # Write records to .avro file
            self.avro_writer.append(data.pop(0))

    def commit(self):
        # The commit is where the upload is actually done for BigQuery (special case).
        # The _create_or_update_table method can be called multiple times;
        # each time, data is appended to the .avro file. When "committing",
        # this .avro file is uploaded and, depending on the load strategy, used.
        if not hasattr(self, "avro_file_name"):
            # There was no data ever uploaded
            # Do nothing
            self.log.info("Nothing to upload!")
            return

        # Clean up after yourself first
        self.avro_writer.close()

        # Fetch the relevant configuration
        project_id = self.table_creation_config.get("database_name", self.database_name)
        assert project_id, "Missing Project ID!"
        load_strategy = self.table_creation_config["load_strategy"]
        primary_key = self.table_creation_config["primary_key"]
        schema_name = self.table_creation_config["schema_name"]
        schema_suffix = self.table_creation_config["schema_suffix"]
        table_name_final = self.table_creation_config["table_name"]
        table_suffix = "__ewah_tmp"

        columns_definition = self.table_creation_config["columns_definition"]
        new_schema_name = schema_name + schema_suffix

        is_full_refresh = (
            load_strategy == EC.LS_INSERT_REPLACE
            or not self.test_if_table_exists(
                table_name=table_name_final,
                schema_name=new_schema_name,
                project_id=project_id,
            )
        )

        conn = self.dwh_hook.dbconn
        ds_new = conn.get_dataset(new_schema_name)

        # Create temp table with .avro file
        if is_full_refresh:
            # temp table is also the final table for full refresh!
            table_name = table_name_final
        else:
            table_name = table_name_final + table_suffix

        # Drop temp table if it already exists
        if self.test_if_table_exists(
            table_name=table_name,
            schema_name=new_schema_name,
            project_id=project_id,
        ):
            # Drop table before re-creating it
            conn.delete_table(
                conn.get_table(TableReference(dataset_ref=ds_new, table_id=table_name))
            )
        # Create temp table with .avro file
        table_obj = Table(".".join([project_id, new_schema_name, table_name]))
        if is_full_refresh and self.partition_field:
            table_obj.time_partitioning = bigquery.TimePartitioning(
                type_=self.partition_type,
                field=self.partition_field,
            )
            if self.require_partition_filter:
                table_obj.require_partition_filter = True
        self.log.info("Uploading data into table now...")
        with open(self.avro_file_name, "rb") as source_file:
            job = conn.load_table_from_file(
                file_obj=source_file,
                destination=table_obj,
                job_id_prefix="ewah_",
                rewind=True,
                job_config=LoadJobConfig(
                    autodetect=False,
                    source_format="AVRO",
                    schema=[
                        SchemaField(name=name, field_type=field["data_type"])
                        for name, field in columns_definition.items()
                    ],
                ),
            )
            try:
                job.result()
            except:
                self.log.info("Errors occured - job errors: {0}".format(job.errors))
                raise
            assert job.state == "DONE", "Invalid job state: {0}".format(job.state)

        if not is_full_refresh:
            # Need to merge new rows into the existing table

            fields_pk = set(primary_key or [])
            fields_all = set(columns_definition.keys() or [])
            fields_non_pk = fields_all - fields_pk

            if load_strategy == EC.LS_UPSERT:
                assert fields_pk
            elif load_strategy == EC.LS_INSERT_ADD:
                fields_pk = []  # Ignore if set
            else:
                raise Exception("Not implemented!")

            merge_statement = """
                MERGE INTO `{target}` AS TARGET
                USING `{source}` AS SOURCE
                ON {condition}

                WHEN MATCHED THEN
                    UPDATE SET {update_fields}

                WHEN NOT MATCHED THEN
                    INSERT ({insert_fields})
                    VALUES ({insert_fields})
            """.format(
                target=".".join([project_id, new_schema_name, table_name_final]),
                source=".".join([project_id, new_schema_name, table_name]),
                condition=" AND ".join(
                    ["TARGET.`{0}` = SOURCE.`{0}`".format(field) for field in fields_pk]
                )
                or "FALSE",
                insert_fields="`{0}`".format("`, `".join(fields_all)),
                update_fields=", ".join(
                    ["`{0}` = SOURCE.`{0}`".format(field) for field in fields_non_pk]
                ),
            )

            self.log.info("Executing query:\n\n{0}\n\n".format(merge_statement))
            job = conn.query(
                query=merge_statement,
                job_id_prefix="ewah_",
            )
            try:
                job.result()
            except:
                self.log.info("Errors occured - job errors: {0}".format(job.errors))
                raise
            assert job.state == "DONE", "Invalid job state: {0}".format(job.state)

            # Remove old temp table from dataset
            conn.delete_table(
                conn.get_table(TableReference(dataset_ref=ds_new, table_id=table_name))
            )

        self.log.info("Done!")



from airflow.models import BaseOperator

from ewah.cleaner import EWAHCleaner
from ewah.constants import EWAHConstants as EC
from ewah.utils.airflow_utils import (
    datetime_utcnow_with_tz,
    airflow_datetime_adjustments as ada,
)
from ewah.hooks.base import EWAHBaseHook
from ewah.uploaders import get_uploader

from datetime import datetime, timedelta
from typing import Optional, List, Dict, Union

import copy
import hashlib
import sys
import time


class EWAHBaseOperator(BaseOperator):
    """Extension of airflow's native Base Operator.

    EWAH operators always work in the same way: They extract raw data from a
    source and load it into a relational database aka data warehouse aka DWH.
    Transformations only happen insofar as they are required to bring data
    into a relational format.

    *How to use*
    The child class extracts data from the source. At the end of the child
    class' execute() function, it calls self.update(). Optional: The child can
    also repeatedly call the self.update() class for chunking.

    *Arguments of self.update():*
    Data is a list of dictionaries, where each list element will become
    one row of data and the contained dict is fieldname:value. Value can be
    None, or a fieldname can be missing in a particular dictionary, in which
    case the value is assumed to be None.

    Implemented DWH engines:
    - PostgreSQL
    - Snowflake

    Next up:
    - BigQuery
    - Redshift
    """

    # Overwrite _NAMES with a list of names that work as aliases when defining
    # DAGs via YAML
    _NAMES = []

    # in child classes, don't overwrite this but add values within __init__!
    template_fields = (
        "load_data_from",
        "load_data_until",
    )

    # Child class must update or overwrite these values
    # A missing element is interpreted as False
    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: False,
    }

    _CONN_TYPE = None  # overwrite me with the required connection type, if applicable

    _INDEX_QUERY = """
        CREATE INDEX IF NOT EXISTS {0}
        ON "{1}"."{2}" ({3})
    """

    _metadata = {}  # to be updated by operator, if applicable

    def __init__(self, *args, **kwargs):
        try:
            dag = kwargs.get("dag")  # For error handling below
            return self.base_init(*args, **kwargs)
        except Exception as e:
            # Add useful information to the raised Exception as it will be displayed
            # in the Airflow UI
            exc_type, exc_obj, exc_tb = sys.exc_info()
            try:
                dag_id = (dag or self.dag).dag_id
            except Exception as ex:
                dag_id = "[unknown]"
            raise Exception(
                """

            EWAH Failed to import the DAG.

            \tDAG: {3}
            \tTask: {4}
            \tOperator: {2}

            \tException Type: {0}
            \tError: {1}

            See below for positional and keyword: {5}

            """.replace(
                    "            ", ""
                ).format(
                    exc_type,
                    str(e),
                    type(self).__name__,
                    dag_id,
                    kwargs["task_id"],
                    "\n\n\tPositional arguments:\n\t\t"
                    + "\n\t\t".join(args or ["(None)"])
                    + "\n\n\tKeyword arguments:\n\t\t"
                    + "\n\t\t".join(
                        ["{0}: {1}".format(k, str(v)) for (k, v) in kwargs.items()]
                        or ["(None)"]
                    ),
                )
            )

    def base_init(
        self,
        source_conn_id,
        dwh_engine,
        dwh_conn_id,
        extract_strategy,
        load_strategy,
        target_table_name,
        target_schema_name,
        target_schema_suffix="_next",
        target_database_name=None,  # Only for Snowflake
        load_data_from=None,  # set a minimum date e.g. for reloading of data
        reload_data_from=None,  # load data from this date for new tables
        load_data_from_relative=None,  # optional timedelta for incremental
        load_data_until=None,  # set a maximum date
        load_data_until_relative=None,  # optional timedelta for incremental
        load_data_chunking_timedelta=None,  # optional timedelta to chunk by
        primary_key=None,  # either string or list of strings
        include_columns=None,  # list of columns to explicitly include
        exclude_columns=None,  # list of columns to exclude
        index_columns=[],  # list of columns to create an index on. can be
        # an expression, must be quoted in list if quoting is required.
        hash_columns=None,  # str or list of str - columns to hash pre-upload
        hash_salt=None,  # string salt part for hashing
        wait_for_seconds=120,  # seconds past data_interval_end to wait until
        # wait_for_seconds only applies for incremental loads
        add_metadata=True,
        rename_columns: Optional[Dict[str, str]] = None,  # Rename columns
        subsequent_field=None,  # field name to use for subsequent extract strategy
        default_timezone=None,  # specify a default time zone for tz-naive datetimes
        use_temp_pickling=True,  # use new upload method if True - use it by default
        pickling_upload_chunk_size=100000,  # default chunk size for pickle upload
        pickle_compression=None,  # data compression algorithm to use for pickles
        default_values=None,  # dict with default values for columns (to avoid nulls)
        cleaner_class=EWAHCleaner,
        cleaner_callables=None,  # callables or list of callables to run during cleaning
        uploader_class=None,  # Future: deprecate dwh_engine and use this kwarg instead
        additional_uploader_kwargs=None,
        deduplication_before_upload=False,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        if default_values:
            assert isinstance(default_values, dict)

        assert pickle_compression is None or pickle_compression in (
            "gzip",
            "bz2",
            "lzma",
        )

        assert isinstance(rename_columns, (type(None), dict))

        if default_timezone:
            assert dwh_engine in (EC.DWH_ENGINE_POSTGRES,)  # Only for PostgreSQL so far
            assert not ";" in default_timezone  # Avoid SQL Injection

        # Check if the extract and load strategies are allowed in combination
        # Also check if required params are supplied for the strategies
        if extract_strategy == EC.ES_SUBSEQUENT:
            assert subsequent_field
            assert load_strategy in (
                EC.LS_UPSERT,
                EC.LS_INSERT_ADD,
                EC.LS_INSERT_REPLACE,
            )
            # insert_delete makes no sense in the subsequent context
        elif extract_strategy == EC.ES_FULL_REFRESH:
            assert load_strategy in (EC.LS_INSERT_REPLACE, EC.LS_INSERT_ADD)
            # upsert makes no sense - it's a full refresh!
            # insert_delete makes no sense - what to delete? Well, everyting!
        elif extract_strategy == EC.ES_INCREMENTAL:
            assert load_strategy in (EC.LS_UPSERT, EC.LS_INSERT_ADD)
            # replace makes no sense - it's incremental loading after all!
            # insert_delete makes sense but is not yet implemented
        else:
            raise Exception("Invalid extract_strategy {0}!".format(extract_strategy))

        if load_strategy == EC.LS_UPSERT:
            # upserts require a (composite) primary key of some sort
            if not primary_key:
                raise Exception(
                    "If the load strategy is upsert, name of the primary"
                    " key(s) (primary_key) is required!"
                )
        elif load_strategy in (EC.LS_INSERT_ADD, EC.LS_INSERT_REPLACE):
            pass  # No requirements
        else:
            raise Exception("Invalid load_strategy {0}!".format(load_strategy))

        for item in EWAHBaseOperator.template_fields:
            # Make sure template_fields was not overwritten
            _msg = "Operator must not overwrite template_fields!"
            assert item in self.template_fields, _msg

        _msg = 'param "wait_for_seconds" must be a nonnegative integer!'
        assert isinstance(wait_for_seconds, int) and wait_for_seconds >= 0, _msg
        _msg = "extract_strategy {0} not accepted for this operator!".format(
            extract_strategy,
        )
        assert self._ACCEPTED_EXTRACT_STRATEGIES.get(extract_strategy), _msg

        if isinstance(primary_key, str):
            primary_key = [primary_key]

        if isinstance(hash_columns, str):
            hash_columns = [hash_columns]

        if isinstance(include_columns, str):
            include_columns = [include_columns]

        if include_columns and primary_key:
            for col in primary_key:
                if not col in include_columns:
                    _msg = """
                        Primary key {0} is not in the include_columns list.
                        Make sure all primary keys are included.
                        """.format(
                        col
                    )
                    raise Exception(_msg)

        if exclude_columns and isinstance(exclude_columns, str):
            exclude_columns = [exclude_columns]

        if include_columns and exclude_columns:
            _msg = "Don't use include and exclude columns config at the same time!"
            raise Exception(_msg)

        if not dwh_engine or not dwh_engine in EC.DWH_ENGINES:
            _msg = "Invalid DWH Engine: {0}\n\nAccepted Engines:\n\t{1}".format(
                str(dwh_engine),
                "\n\t".join(EC.DWH_ENGINES),
            )
            raise Exception(_msg)

        if index_columns and not dwh_engine == EC.DWH_ENGINE_POSTGRES:
            raise Exception("Indices are only allowed for PostgreSQL DWHs!")

        if (not dwh_engine == EC.DWH_ENGINE_SNOWFLAKE) and target_database_name:
            raise Exception('Received argument for "target_database_name"!')

        _msg = "load_data_from_relative and load_data_until_relative must be"
        _msg += " timedelta if supplied!"
        assert isinstance(
            load_data_from_relative,
            (type(None), timedelta),
        ), _msg
        assert isinstance(
            load_data_until_relative,
            (type(None), timedelta),
        ), _msg
        _msg = "load_data_chunking_timedelta must be timedelta!"
        assert isinstance(
            load_data_chunking_timedelta,
            (type(None), timedelta),
        ), _msg

        if callable(cleaner_callables):
            cleaner_callables = [cleaner_callables]

        self.source_conn_id = source_conn_id
        self.dwh_engine = dwh_engine
        self.dwh_conn_id = dwh_conn_id
        self.extract_strategy = extract_strategy
        self.load_strategy = load_strategy
        self.target_table_name = target_table_name
        self.target_schema_name = target_schema_name
        self.target_schema_suffix = target_schema_suffix
        self.target_database_name = target_database_name
        self.load_data_from = load_data_from
        self.reload_data_from = reload_data_from
        self.load_data_from_relative = load_data_from_relative
        self.load_data_until = load_data_until
        self.load_data_until_relative = load_data_until_relative
        self.load_data_chunking_timedelta = load_data_chunking_timedelta
        self.primary_key = primary_key  # may be used ...
        #   ... by a child class at execution!
        self.include_columns = include_columns
        self.exclude_columns = exclude_columns
        self.index_columns = index_columns
        self.hash_columns = hash_columns
        self.hash_salt = hash_salt
        self.wait_for_seconds = wait_for_seconds
        self.add_metadata = add_metadata
        self.rename_columns = rename_columns
        self.subsequent_field = subsequent_field
        self.default_timezone = default_timezone
        self.use_temp_pickling = use_temp_pickling
        self.pickling_upload_chunk_size = pickling_upload_chunk_size
        self.pickle_compression = pickle_compression
        self.default_values = default_values
        self.cleaner_class = cleaner_class
        self.cleaner_callables = cleaner_callables
        self.deduplication_before_upload = deduplication_before_upload

        self.uploader_class = uploader_class or get_uploader(self.dwh_engine)
        self.additional_uploader_kwargs = additional_uploader_kwargs or {}

        _msg = "DWH hook does not support extract strategy {0}!".format(
            extract_strategy,
        )
        # assert self.uploader._ACCEPTED_EXTRACT_STRATEGIES.get(extract_strategy), _msg

    def ewah_execute(self, context):
        raise Exception("You need to overwrite me!")

    def execute(self, context):
        """Why this method is defined here:
        When executing a task, airflow calls this method. Generally, this
        method contains the "business logic" of the individual operator.
        However, EWAH may want to do some actions for all operators. Thus,
        the child operators shall have an ewah_execute() function which is
        called by this general execute() method.
        """

        self.log.info(
            """

            Running EWAH Operator {0}.
            DWH: {1} (connection id: {2})
            Extract Strategy: {3}
            Load Strategy: {4}

            """.format(
                str(self),
                self.dwh_engine,
                self.dwh_conn_id,
                self.extract_strategy,
                self.load_strategy,
            )
        )

        # required for metadata in data upload
        self._execution_time = datetime_utcnow_with_tz()
        self._context = context

        cleaner_callables = self.cleaner_callables or []

        if self.source_conn_id:
            # resolve conn id here & delete the object to avoid usage elsewhere
            self.source_conn = EWAHBaseHook.get_connection(self.source_conn_id)
            self.source_hook = self.source_conn.get_hook()
            if callable(getattr(self.source_hook, "get_cleaner_callables", None)):
                hook_callables = self.source_hook.get_cleaner_callables()
                if callable(hook_callables):
                    cleaner_callables.append(hook_callables)
                elif hook_callables:
                    # Ought to be list of callables
                    cleaner_callables += hook_callables
        del self.source_conn_id

        if self._CONN_TYPE:
            assert (
                self._CONN_TYPE == self.source_conn.conn_type
            ), "Error - connection type must be {0}!".format(self._CONN_TYPE)

        uploader_callables = self.uploader_class.get_cleaner_callables()
        if callable(uploader_callables):
            cleaner_callables.append(uploader_callables)
        elif uploader_callables:
            cleaner_callables += uploader_callables

        self.uploader = self.uploader_class(
            dwh_conn=EWAHBaseHook.get_connection(self.dwh_conn_id),
            cleaner=self.cleaner_class(
                default_row=self.default_values,
                include_columns=self.include_columns,
                exclude_columns=self.exclude_columns,
                add_metadata=self.add_metadata,
                rename_columns=self.rename_columns,
                hash_columns=self.hash_columns,
                hash_salt=self.hash_salt,
                additional_callables=cleaner_callables,
            ),
            table_name=self.target_table_name,
            schema_name=self.target_schema_name,
            schema_suffix=self.target_schema_suffix,
            database_name=self.target_database_name,
            primary_key=self.primary_key,
            load_strategy=self.load_strategy,
            use_temp_pickling=self.use_temp_pickling,
            pickling_upload_chunk_size=self.pickling_upload_chunk_size,
            pickle_compression=self.pickle_compression,
            deduplication_before_upload=self.deduplication_before_upload,
            **self.additional_uploader_kwargs,
        )

        # If applicable: set the session's default time zone
        if self.default_timezone:
            self.uploader.dwh_hook.execute(
                "SET timezone TO '{0}'".format(self.default_timezone)
            )

        # Create a new copy of the target table.
        # This is so data is loaded into a new table and if data loading
        # fails, the original data is not corrupted. At a new try or re-run,
        # the original table is just copied anew.
        if not self.load_strategy == EC.LS_INSERT_REPLACE:
            # insert_replace always drops and replaces the tables completely
            self.uploader.copy_table()

        # set load_data_from and load_data_until as required
        data_from = ada(self.load_data_from)
        data_until = ada(self.load_data_until)
        if self.extract_strategy == EC.ES_INCREMENTAL:
            _tdz = timedelta(days=0)  # aka timedelta zero
            _ed = context["data_interval_start"]
            _ned = context["data_interval_end"]

            # normal incremental load
            _ed -= self.load_data_from_relative or _tdz
            data_from = min(_ed, data_from or _ed)
            if not self.test_if_target_table_exists():
                # Load data from scratch!
                data_from = ada(self.reload_data_from) or data_from

            _ned += self.load_data_until_relative or _tdz
            data_until = max(_ned, data_until or _ned)

        elif self.extract_strategy in (EC.ES_FULL_REFRESH, EC.ES_SUBSEQUENT):
            # Values may still be set as static values
            data_from = ada(self.reload_data_from) or data_from

        else:
            _msg = "Must define load_data_from etc. behavior for load strategy!"
            raise Exception(_msg)

        self.data_from = data_from
        self.data_until = data_until
        # del variables to make sure they are not used later on
        del self.load_data_from
        del self.reload_data_from
        del self.load_data_until
        del self.load_data_until_relative
        if not self.extract_strategy == EC.ES_SUBSEQUENT:
            # keep this param for subsequent loads
            del self.load_data_from_relative

        # Have an option to wait until a short period (e.g. 2 minutes) past
        # the incremental loading range timeframe to ensure that all data is
        # loaded, useful e.g. if APIs lag or if server timestamps are not
        # perfectly accurate.
        # When a DAG is executed as soon as possible, some data sources
        # may not immediately have up to date data from their API.
        # E.g. querying all data until 12.30pm only gives all relevant data
        # after 12.32pm due to some internal delays. In those cases, make
        # sure the (incremental loading) DAGs don't execute too quickly.
        if self.wait_for_seconds and self.extract_strategy == EC.ES_INCREMENTAL:
            wait_until = context.get("data_interval_end")
            if wait_until:
                wait_until += timedelta(seconds=self.wait_for_seconds)
                self.log.info(
                    "Awaiting execution until {0}...".format(
                        str(wait_until),
                    )
                )
            while wait_until and datetime_utcnow_with_tz() < wait_until:
                # Only sleep a maximum of 5s at a time
                wait_for_timedelta = wait_until - datetime_utcnow_with_tz()
                time.sleep(max(0, min(wait_for_timedelta.total_seconds(), 5)))

        # execute operator
        if self.load_data_chunking_timedelta and data_from and data_until:
            # Chunking to avoid OOM
            assert data_until > data_from
            assert self.load_data_chunking_timedelta > timedelta(days=0)
            while self.data_from < data_until:
                self.data_until = min(
                    self.data_from + self.load_data_chunking_timedelta, data_until
                )
                self.log.info(
                    "Now loading from {0} to {1}...".format(
                        str(self.data_from), str(self.data_until)
                    )
                )
                self.ewah_execute(context)
                self.data_from += self.load_data_chunking_timedelta
        else:
            self.ewah_execute(context)

        # Run final scripts
        # TODO: Include indexes into uploader and then remove this step
        self.uploader.finalize_upload()

        # if PostgreSQL and arg given: create indices
        for column in self.index_columns:
            assert self.dwh_engine == EC.DWH_ENGINE_POSTGRES
            # Use hashlib to create a unique 63 character string as index
            # name to avoid breaching index name length limits & accidental
            # duplicates / missing indices due to name truncation leading to
            # identical index names.
            self.uploader.dwh_hook.execute(
                self._INDEX_QUERY.format(
                    "__ewah_"
                    + hashlib.blake2b(
                        (
                            self.target_schema_name
                            + self.target_schema_suffix
                            + "."
                            + self.target_table_name
                            + "."
                            + column
                        ).encode(),
                        digest_size=28,
                    ).hexdigest(),
                    self.target_schema_name + self.target_schema_suffix,
                    self.target_table_name,
                    column,
                )
            )

        # commit only at the end, so that no data may be committed before an
        # error occurs.
        self.log.info("Now committing changes!")
        self.uploader.commit()
        self.uploader.close()

    def test_if_target_table_exists(self):
        # TODO: move this function to uploader
        # Need to use existing hook to work within open transaction
        kwargs = {
            "table_name": self.target_table_name,
            "schema_name": self.target_schema_name + self.target_schema_suffix,
        }
        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            kwargs["database_name"] = self.target_database_name
            return self.uploader.test_if_table_exists(**kwargs)
        if self.dwh_engine == EC.DWH_ENGINE_BIGQUERY:
            kwargs["project_id"] = self.target_database_name
            return self.uploader.test_if_table_exists(**kwargs)
        if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
            return self.uploader.test_if_table_exists(**kwargs)
        # For a new DWH, need to manually check if function works properly
        # Thus, fail until explicitly added
        raise Exception("Function not implemented!")

    def get_max_value_of_column(self, column_name):
        # Deprecated - better to call uploader directly if able
        return self.uploader.get_max_value_of_column(column_name=column_name)

    def upload_data(self, data=None):
        """Upload data, no matter the source. Call this functions in the child
        operator whenever data is available for upload, as often as needed.
        """
        if not data:
            self.log.info("No data to upload!")
            return

        if self.add_metadata:
            metadata = copy.deepcopy(self._metadata)  # from individual operator
            interval_start = self._context["data_interval_start"]
            interval_start = datetime.fromtimestamp(
                interval_start.timestamp(), interval_start.tz
            )
            interval_end = self._context["data_interval_end"]
            interval_end = datetime.fromtimestamp(
                interval_end.timestamp(), interval_end.tz
            )
            metadata.update(
                {
                    "_ewah_extract_strategy": self.extract_strategy,
                    "_ewah_load_strategy": self.load_strategy,
                    "_ewah_executed_at": self._execution_time,
                    "_ewah_dag_id": self._context["dag"].dag_id,
                    "_ewah_dag_run_id": self._context["run_id"],
                    "_ewah_dag_run_data_interval_start": interval_start,
                    "_ewah_dag_run_data_interval_end": interval_end,
                }
            )
        else:
            metadata = None

        return self.uploader.upload_data(data, metadata)



"""
Airbyte's implementation of this connector has been an important source of information
in the process of building this EWAH connector. Check out Airbyte for more info!
"""

from ewah.hooks.base import EWAHBaseHook

from datetime import datetime
import urllib.parse
import pendulum
import hashlib
import hmac
import boto3
import requests
import time
import pytz
import copy


class EWAHAmazonSellerCentralHook(EWAHBaseHook):

    _APIS = {
        "orders": {
            "path": "/orders/v0/orders",
            "default_rate": 0.0055,
            "datetime_filter_params": ["CreatedAfter", "LastUpdatedAfter"],
            "datetime_filter_field": "LastUpdateDate",
            "primary_key": "AmazonOrderId",
            "payload_field": "Orders",
            "sideloads": ["orderitems"],
        },
        "orderitems": {  # only for use as sideload
            "path": "/orders/v0/orders/{id}/orderItems",
            "default_rate": 0.0055,
            "payload_field": "OrderItems",
        },
    }

    _THROTTLING_DICT = {}

    _ATTR_RELABEL = {}

    conn_name_attr = "ewah_amazon_seller_central_conn_id"
    default_conn_name = "ewah_amazon_seller_central_default"
    conn_type = "ewah_amazon_seller_central"
    hook_name = "EWAH Amazon Seller Central Connection"

    @staticmethod
    def get_ui_field_behaviour():
        # Hide all and use custom fields only
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login", "password"],
            "relabeling": {},
        }

    @staticmethod
    def get_connection_form_widgets():
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import (
            BS3TextFieldWidget,
            BS3PasswordFieldWidget,
        )
        from wtforms import StringField

        return {
            "extra__ewah_amazon_seller_central__aws_access_key_id": StringField(
                "AWS Access Key ID", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__aws_secret_access_key": StringField(
                "AWS Secret Access Key", widget=BS3PasswordFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__aws_arn_role": StringField(
                "AWS ARN - Role", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__lwa_client_id": StringField(
                "LoginWithAmazon (LWA) Client ID", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__lwa_client_secret": StringField(
                "LoginWithAmazon (LWA) Client Secret", widget=BS3PasswordFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__refresh_token": StringField(
                "Refresh Token", widget=BS3PasswordFieldWidget()
            ),
        }

    @staticmethod
    def get_marketplace_details_tuple(marketplace_region):
        base_url = "https://sellingpartnerapi"
        return {
            "AE": (f"{base_url}-eu.amazon.com", "A2VIGQ35RCS4UG", "eu-west-1"),
            "DE": (f"{base_url}-eu.amazon.com", "A1PA6795UKMFR9", "eu-west-1"),
            "PL": (f"{base_url}-eu.amazon.com", "A1C3SOZRARQ6R3", "eu-west-1"),
            "EG": (f"{base_url}-eu.amazon.com", "ARBP9OOSHTCHU", "eu-west-1"),
            "ES": (f"{base_url}-eu.amazon.com", "A1RKKUPIHCS9HS", "eu-west-1"),
            "FR": (f"{base_url}-eu.amazon.com", "A13V1IB3VIYZZH", "eu-west-1"),
            "IN": (f"{base_url}-eu.amazon.com", "A21TJRUUN4KGV", "eu-west-1"),
            "IT": (f"{base_url}-eu.amazon.com", "APJ6JRA9NG5V4", "eu-west-1"),
            "NL": (f"{base_url}-eu.amazon.com", "A1805IZSGTT6HS", "eu-west-1"),
            "SA": (f"{base_url}-eu.amazon.com", "A17E79C6D8DWNP", "eu-west-1"),
            "SE": (f"{base_url}-eu.amazon.com", "A2NODRKZP88ZB9", "eu-west-1"),
            "TR": (f"{base_url}-eu.amazon.com", "A33AVAJ2PDY3EV", "eu-west-1"),
            "UK": (f"{base_url}-eu.amazon.com", "A1F83G8C2ARO7P", "eu-west-1"),
            "AU": (f"{base_url}-fe.amazon.com", "A39IBJ37TRP1C6", "us-west-2"),
            "JP": (f"{base_url}-fe.amazon.com", "A1VC38T7YXB528", "us-west-2"),
            "SG": (f"{base_url}-fe.amazon.com", "A19VAU5U5O7RUS", "us-west-2"),
            "US": (f"{base_url}-na.amazon.com", "ATVPDKIKX0DER", "us-east-1"),
            "BR": (f"{base_url}-na.amazon.com", "A2Q3Y263D00KWC", "us-east-1"),
            "CA": (f"{base_url}-na.amazon.com", "A2EUQ1WTGCTBG2", "us-east-1"),
            "MX": (f"{base_url}-na.amazon.com", "A1AM78C64UM0Y8", "us-east-1"),
            "GB": (f"{base_url}-eu.amazon.com", "A1F83G8C2ARO7P", "eu-west-1"),
        }[marketplace_region]

    @classmethod
    def get_cleaner_callables(cls):
        def string_to_date(row):
            for key, value in row.items():
                if key.endswith("Date"):
                    row[key] = datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ").replace(
                        tzinfo=pytz.utc
                    )
            return row

        return string_to_date

    @classmethod
    def validate_marketplace_region(cls, marketplace_region, raise_on_failure=False):
        # Validate whether a region exists. Return boolean or raise exception.
        try:
            cls.get_marketplace_details_tuple(marketplace_region)
        except:
            if raise_on_failure:
                raise
            else:
                return False
        finally:
            return True

    @staticmethod
    def _sign_msg(key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    @property
    def access_token(self):
        if (
            not hasattr(self, "_access_token")
            or pendulum.now() > self._access_token_expires_at
        ):
            self.log.info("Requesting new access token...")
            requested_at = pendulum.now()
            url = "https://api.amazon.com/auth/o2/token"
            payload = {
                "grant_type": "refresh_token",
                "client_id": self.conn.lwa_client_id,
                "client_secret": self.conn.lwa_client_secret,
                "refresh_token": self.conn.refresh_token,
            }
            response = requests.post(url, data=payload)
            assert response.status_code == 200, response.text
            response_data = response.json()
            self._access_token = response_data["access_token"]
            self._access_token_expires_at = requested_at.add(
                # add a 15% error margin to avoid expiration during a call
                seconds=(response_data["expires_in"] * 0.85),
            )
            self.log.info(
                "New Amazon oauth access token is valid until {0}...".format(
                    self._access_token_expires_at.isoformat()
                )
            )
        return self._access_token

    @property
    def boto3_role_credentials(self):
        if (
            not hasattr(self, "_boto3_role_credentials")
            or pendulum.now() > self._boto3_session_expires_at
        ):
            self.log.info("Logging in to AWS...")
            boto3_client = boto3.client(
                "sts",
                aws_access_key_id=self.conn.aws_access_key_id,
                aws_secret_access_key=self.conn.aws_secret_access_key,
            )
            role = boto3_client.assume_role(
                RoleArn=self.conn.aws_arn_role, RoleSessionName="guid"
            )
            self._boto3_role_credentials = role["Credentials"]
            # Add a 15% error margin to avoid expiration during a call
            self._boto3_session_expires_at = pendulum.now() + 0.85 * (
                role["Credentials"]["Expiration"] - pendulum.now()
            )
            self.log.info(
                "Logged in to AWS! Login expires at: {0}".format(
                    self._boto3_session_expires_at.isoformat()
                )
            )
        return self._boto3_role_credentials

    @property
    def aws_session_token(self):
        return self.boto3_role_credentials.get("SessionToken")

    @property
    def aws_session_secret_access_key(self):
        return self.boto3_role_credentials.get("SecretAccessKey")

    @property
    def aws_session_access_key_id(self):
        return self.boto3_role_credentials.get("AccessKeyId")

    def make_api_call_and_return_payload(
        self,
        resource,
        path,
        marketplace_region,
        since_date=None,
        since_type=None,
        next_token=None,
        logging=True,
        additional_api_call_params=None,
    ):

        if self._THROTTLING_DICT.get(resource):
            # APIs use token bucket-style requests throttling
            next_call_after = self._THROTTLING_DICT[resource]
            if logging:
                self.log.info(
                    f"Waiting for next call until {next_call_after.isoformat()}..."
                )
            sleeptime = next_call_after - pendulum.now()
            time.sleep(max(0, sleeptime.microseconds / 1000000 + sleeptime.seconds))

        current_ts = pendulum.now("utc")  # as late as possible -> set after waiting
        amz_date = current_ts.strftime("%Y%m%dT%H%M%SZ")
        datestamp = current_ts.strftime("%Y%m%d")

        endpoint, marketplace_id, region = self.get_marketplace_details_tuple(
            marketplace_region
        )
        metadata = self._APIS[resource]
        if not path.startswith("/"):
            path = "/" + path
        if since_date:
            if since_type == "CreatedAfter":
                filter_field = metadata["datetime_filter_params"][0]
            elif since_type == "LastUpdatedAfter":
                filter_field = metadata["datetime_filter_params"][1]
            else:
                raise Exception("This exception ought to have been caught earlier.")

        url = "".join([endpoint, path])

        params = copy.deepcopy(additional_api_call_params or {})
        params["MarketplaceIds"] = marketplace_id
        params["MaxResultsPerPage"] = 100

        if since_date:
            params[filter_field] = (
                since_date.replace(microsecond=0)
                .astimezone(pytz.utc)
                .isoformat()
                .replace("+00:00", "Z")
            )
            if not next_token and logging:
                self.log.info(
                    f"Loading data from {filter_field}: {params[filter_field]}..."
                )

        if next_token:
            params["NextToken"] = next_token

        url_parsed = urllib.parse.urlparse(url)
        uri = urllib.parse.quote(url_parsed.path)
        host = url_parsed.hostname
        service = "execute-api"

        canonical_querystring = "&".join(
            map(
                lambda param: "=".join(param),
                sorted(
                    [
                        [str(key), urllib.parse.quote_plus(str(value))]
                        for key, value in params.items()
                    ]
                ),
            )
        )

        headers_to_sign = {
            "host": host,
            "x-amz-date": amz_date,
            "x-amz-security-token": self.aws_session_token,
        }
        ordered_headers = dict(sorted(headers_to_sign.items(), key=lambda h: h[0]))
        canonical_headers = "".join(
            map(lambda h: ":".join(h) + "\n", ordered_headers.items())
        )
        signed_headers = ";".join(ordered_headers.keys())
        payload_hash = hashlib.sha256("".encode("utf-8")).hexdigest()
        canonical_request = "\n".join(
            [
                "GET",
                uri,
                canonical_querystring,
                canonical_headers,
                signed_headers,
                payload_hash,
            ]
        )

        credential_scope = "/".join([datestamp, region, service, "aws4_request"])
        string_to_sign = "\n".join(
            [
                "AWS4-HMAC-SHA256",
                amz_date,
                credential_scope,
                hashlib.sha256(canonical_request.encode("utf-8")).hexdigest(),
            ]
        )

        datestamp_signed = self._sign_msg(
            ("AWS4" + self.aws_session_secret_access_key).encode("utf-8"), datestamp
        )
        region_signed = self._sign_msg(datestamp_signed, region)
        service_signed = self._sign_msg(region_signed, service)
        aws4_request_signed = self._sign_msg(service_signed, "aws4_request")
        signature = hmac.new(
            aws4_request_signed, string_to_sign.encode("utf-8"), hashlib.sha256
        ).hexdigest()
        authorization_header = (
            "AWS4-HMAC-SHA256 Credential={}/{}, SignedHeaders={}, Signature={}".format(
                self.aws_session_access_key_id,
                credential_scope,
                signed_headers,
                signature,
            )
        )

        headers = {
            "content-type": "application/json",
            "host": host,
            "user-agent": "python-requests",
            "x-amz-access-token": self.access_token,
            "x-amz-date": amz_date,
            "x-amz-security-token": self.aws_session_token,
            "authorization": authorization_header,
        }

        response = requests.get(url, params=params, headers=headers)
        assert response.status_code == 200, response.text

        self._THROTTLING_DICT[resource] = pendulum.now().add(
            seconds=(
                1
                / float(
                    response.headers.get(
                        "x-amzn-RateLimit-Limit",
                        metadata.get("default_rate", 0.0055),
                    )
                )
            )
        )
        return response.json()["payload"]

    def get_data_in_batches(
        self,
        resource,
        marketplace_region,
        path=None,
        since_date=None,
        since_type=None,
        sideloads=None,
        logging=True,
        additional_api_call_params=None,
    ):
        next_token = None
        metadata = self._APIS[resource]
        path = path or metadata["path"]
        if since_date:
            assert since_type in ("CreatedAfter", "LastUpdatedAfter")
        else:
            assert since_type is None

        while True:
            if logging:
                self.log.info("Making new request...")
            response_payload = self.make_api_call_and_return_payload(
                resource,
                path,
                marketplace_region,
                since_date,
                since_type,
                next_token,
                logging,
                additional_api_call_params,
            )
            next_token = response_payload.get("NextToken")

            data = response_payload.pop(metadata["payload_field"], None)
            if data:
                if sideloads:
                    # sideloads are API calls that need to be made per row of original
                    # result. E.g. orderitems for each order. This requires a loop
                    # over each original row and one request for each row.
                    if logging:
                        self.log.info(
                            f"Loading sideloads ({(', '.join(sideloads))}) for "
                            f"{len(data)} rows of {resource}."
                        )
                    for datum in data:
                        id = datum[metadata["primary_key"]]
                        for sideload in sideloads:
                            # sideloads have no "since_date"
                            sideload_data = []
                            sideload_meta = self._APIS[sideload]
                            for batch in self.get_data_in_batches(
                                resource=sideload,
                                path=sideload_meta["path"].format(id=id),
                                marketplace_region=marketplace_region,
                                logging=False,
                            ):
                                if batch:
                                    sideload_data += batch
                            datum["sideload_" + sideload] = sideload_data
                yield data
            elif next_token:
                raise Exception(
                    "No data was returned, but a NextToken -> something went wrong!"
                )
            if not next_token:
                # We've reached the last page
                break



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator
from ewah.hooks.amazon_seller_central import EWAHAmazonSellerCentralHook

from datetime import datetime


class EWAHAmazonSellerCentralOperator(EWAHBaseOperator):

    _NAMES = ["asc", "seller_central", "amazon_seller_central", "sp-api"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: True,
    }

    def __init__(
        self,
        marketplace_region,
        resource=None,
        sideloads=None,
        additional_api_call_params=None,
        *args,
        **kwargs,
    ):
        self.resource = resource or kwargs["target_table_name"]
        self.marketplace_region = marketplace_region
        if sideloads:
            if isinstance(sideloads, str):
                self.sideloads = [sideloads]
            else:
                assert isinstance(sideloads, list), "sideloads must be str or list!"
                self.sideloads = sideloads
        else:
            self.sideloads = None

        assert EWAHAmazonSellerCentralHook.validate_marketplace_region(
            marketplace_region
        ), f"Marketplace Region {marketplace_region} is invalid!"

        metadata = EWAHAmazonSellerCentralHook._APIS[self.resource]
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = metadata["datetime_filter_field"]
        kwargs["primary_key"] = metadata["primary_key"]

        assert additional_api_call_params is None or isinstance(
            additional_api_call_params, dict
        ), "additional_api_call_params must be a dict!"
        self.additional_api_call_params = additional_api_call_params

        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            since_type = "LastUpdatedAfter"
        else:
            # Reload data from scratch
            data_from = self.data_from
            assert (
                data_from
            ), "If reloading data from scratch, the reload_data_from kwarg must be set!"
            since_type = "CreatedAfter"

        for batch in self.source_hook.get_data_in_batches(
            resource=self.resource,
            marketplace_region=self.marketplace_region,
            since_date=data_from,
            since_type=since_type,
            sideloads=self.sideloads,
            additional_api_call_params=self.additional_api_call_params,
        ):
            self.upload_data(batch)



from ewah.hooks.base import EWAHBaseHook

import requests

from datetime import datetime, date


class EWAHPersonioHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "client_id": "login",
        "client_secret": "password",
    }

    conn_name_attr = "ewah_personio_conn_id"
    default_conn_name = "ewah_personio_default"
    conn_type = "ewah_personio"
    hook_name = "EWAH Personio Connection"

    BASE_URL = "https://api.personio.de/v1/"

    ENDPOINTS = {
        "employees": "company/employees",
        "absences": "company/time-offs",
        "time-offs": "company/time-offs",
        "projects": "company/attendances/projects",
        "attendances": "company/attendances",
    }

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "host"],
            "relabeling": {
                "login": "Client ID",
                "password": "Client Secret",
            },
        }

    @classmethod
    def validate_resource(cls, resource):
        assert (
            resource in cls.ENDPOINTS.keys()
        ), "'{0}' is not a valid resource! Valid resources: {1}".format(
            resource, ", ".join(cls.ENDPOINTS.keys())
        )
        return True

    @property
    def token(self):
        # Token needs to be re-requested for every API call!
        token_request = requests.post(
            self.BASE_URL + "auth",
            headers={"Accept": "application/json", "X-Personio-Partner-ID": "ewah"},
            params={
                "client_id": self.conn.client_id,
                "client_secret": self.conn.client_secret,
            },
        )
        assert token_request.status_code == 200, token_request.text
        token_data = token_request.json()
        assert token_data.get("success"), token_data
        return token_data["data"]["token"]

    def get_data_in_batches(self, resource, data_from=None):
        self.validate_resource(resource)
        url = self.BASE_URL + self.ENDPOINTS[resource]
        headers = {
            "Accept": "application/json",
            "X-Personio-Partner-ID": "ewah",
        }
        params = {
            "limit": 200,
            "offset": 0,
        }
        if resource == "attendances":
            # These params are required. Just make them ridiculous.
            params["start_date"] = "1900-01-01"
            params["end_date"] = "2100-01-01"
            if data_from:
                if isinstance(data_from, (date, datetime)):
                    params["updated_from"] = data_from.isoformat()
                else:
                    params["updated_from"] = data_from
        else:
            assert not data_from, "data_from is only valid for attendances!"
        if resource in ["absences", "time-offs"]:
            # It appears as if the "limit" parameter is used like a "page"
            # parameter in Personio's API for absences. Hence, start with one,
            # and incremental like a serial.
            params["offset"] = 1

        while True:
            self.log.info("Requesting a page of data...")
            # Token needs to be re-requested for every API call!
            headers["Authorization"] = "Bearer {0}".format(self.token)
            response = requests.get(url, params=params, headers=headers)
            assert response.status_code == 200, response.text
            response_data = response.json()
            assert response_data.get("success"), response_data
            if response_data.get("data"):
                # Unpack attributes dict
                # Format of dict differs across endpoints! Cover all of them
                data = response_data.pop("data")
                for datum in data:
                    attributes = datum.pop("attributes", {})
                    for attribute, value in attributes.items():
                        if resource == "employees":
                            value = value["value"]
                        if (
                            resource in ["absences", "time-offs"]
                            and attribute == "employee"
                        ):
                            # Don't pull PII from absences, only employee ID!
                            # If able, that data can be pulled from the employees
                            # endpoint!
                            value = value["attributes"]["id"]["value"]
                        datum[attribute] = value
                data_len = len(data)
                yield data
            else:
                break
            if (  # don't use the page parameter - doesn't work properly!
                response_data.get("limit") and data_len < int(response_data["limit"])
            ) or (not response_data.get("limit")):
                # If there's not limit parameter, then there's no pagination at all
                # We're done here
                break
            if resource in ["absences", "time-offs"]:
                # As discussed above: in the case of absences/time-offs (same endpoint),
                # the "offset" parameter appears to be used like a "page" parameter.
                # This is a workaround to work with this buggy API behavior.
                params["offset"] = params["offset"] + 1
            else:
                params["offset"] = int(response_data["offset"]) + int(
                    response_data["limit"]
                )
